import { IDigraphEdge, IDigraph } from '@algorithm.ts/graph.types';
import { IPriorityQueue } from '@algorithm.ts/queue';

/**
 * Make an object deep-readonly.
 */
type DeepReadonly<T extends object> = {
    readonly [P in keyof T]: T[P] extends object ? DeepReadonly<T[P]> : Readonly<T[P]>;
};

interface IDijkstraEdge<C extends number | bigint> extends IDigraphEdge {
    /**
     * The cost of walking along this side.
     */
    cost: C;
}
interface IDijkstraGraph<C extends number | bigint> extends IDigraph<IDijkstraEdge<C>> {
    /**
     * The source node. (0-index)
     */
    readonly source: number;
}
interface IDijkstraProps<C extends number | bigint> {
    /**
     * The value represent the zero cost, 0 for number and 0n for bigint.
     */
    readonly ZERO: C;
    /**
     * A big number / bigint, representing the unreachable cost.
     */
    readonly INF: C;
}
interface IDijkstraOptions<C extends number | bigint> {
    /**
     * A big number / bigint, representing the unreachable cost.
     */
    readonly INF?: C;
}
interface IDijkstraResult<C extends number | bigint> {
    /**
     * A big number, representing the unreachable cost.
     */
    readonly INF: C;
    /**
     * Source point
     */
    readonly source: number;
    /**
     * Record the shortest path parent source point to the specified point.
     * For example: bestFrom[x] represents the previous position of x in the shortest path
     *              parent the source point to x.
     */
    readonly bestFrom: ReadonlyArray<number>;
    /**
     * An array recording the shortest distance to the source point.
     */
    readonly dist: ReadonlyArray<C>;
}

declare const dijkstra: (graph: DeepReadonly<IDijkstraGraph<number>>, options?: IDijkstraOptions<number>) => IDijkstraResult<number>;
declare const dijkstraBigint: (graph: DeepReadonly<IDijkstraGraph<bigint>>, options?: IDijkstraOptions<bigint>) => IDijkstraResult<bigint>;

interface IStateNode<C extends number | bigint> {
    readonly pos: number;
    readonly cost: C;
}
/**
 * The dijkstra algorithm, optimized with priority queue.
 *
 * !!!NOTE dijkstra cannot work with negative cycle.
 *
 * @param graph
 * @param options
 * @see https://me.guanghechen.com/post/algorithm/graph/shortest-path/dijkstra
 */
declare class Dijkstra<C extends number | bigint> {
    protected readonly ZERO: C;
    protected readonly INF: C;
    protected readonly dist: C[];
    protected readonly bestFrom: number[];
    protected readonly done: boolean[];
    protected readonly Q: IPriorityQueue<IStateNode<C>>;
    constructor(props: IDijkstraProps<C>);
    dijkstra(graph: DeepReadonly<IDijkstraGraph<C>>, options?: IDijkstraOptions<C>): IDijkstraResult<C>;
}

export { Dijkstra, type IDijkstraEdge, type IDijkstraGraph, type IDijkstraOptions, type IDijkstraProps, type IDijkstraResult, dijkstra as default, dijkstra, dijkstraBigint };
